package net.cephas.dwr;
import java.util.HashMap;
import java.util.Map;

import uk.ltd.getahead.dwr.impl.DefaultConfiguration;
import uk.ltd.getahead.dwr.impl.SignatureParser;
import uk.ltd.getahead.dwr.util.Logger;

/**
 * Uses the FluentInterface described by Martin Fowler on his blog:<p>
 * 
 * 	<a href="http://www.martinfowler.com/bliki/FluentInterface.html">http://www.martinfowler.com/bliki/FluentInterface.html</a><p>
 * 
 * to wire up the configuration programatically rather than having
 * to use dwr.xml. In order to use this style, you'll need to:
 * 	<ul>
 * 		<li> create a class that extends DWRServlet and use that class as your DWR servlet
 * 		<li> add a configuration param in web.xml called uk.ltd.getahead.dwr.Configuration
 * 		and set the value to net.cephas.dwr.FluentConfiguration
 * 		<li> add a configuration param in web.xml called skipDefaultConfig and set the
 * 		value to true
 * 		<pre>
 * &lt;servlet&gt;
 *   &lt;servlet-name&gt;dwr&lt;/servlet-name&gt;
 * 	 &lt;servlet-class&gt;net.cephas.dwr.FluentDWRServlet&lt;/servlet-class&gt;
 * 	 &lt;init-param&gt;
 * 	   &lt;param-name>uk.ltd.getahead.dwr.Configuration&lt;/param-name&gt;
 * 	   &lt;param-value&gt;net.cephas.dwr.FluentConfiguration&lt;/param-value&gt;
 * 	 &lt;/init-param&gt;
 * 	 &lt;init-param&gt;
 * 	   &lt;param-name>skipDefaultConfig&lt;/param-name&gt;
 * 	   &lt;param-value&gt;true&lt;/param-value&gt;
 * 	 &lt;/init-param&gt;
 * &lt;/servlet&gt;
 * 		</pre>
 * 		<li> and then override the {@link uk.ltd.getahead.dwr.AbstractDWRServlet#configure}
 * 		method in the servlet and use the fluent style of configuration which looks something
 * 		like this:
 *  		<pre>
 * public void configure(ServletConfig config, Configuration configuration) throws ServletException {
 *   if (configuration instanceof FluentConfiguration) {
 *  	   FluentConfiguration fluentconfig = (FluentConfiguration)configuration;
 *  	   fluentconfig
 *	     .withConverterType("dog", "com.yourcompany.beans.Dog")
 *	    	 .withCreatorType("ejb", "com.yourcompany.dwr.creator.EJBCreator")
 *	    	 .withCreator("new", "ApartmentDAO")
 *		   .addParam("scope", "session")
 *	    	   .addParam("class", "com.yourcompany.dao.ApartmentDAO")
 *	    	   .include("getApartments")
 *	    	   .exclude("saveApartment")
 *	       .withAuth("method", "role")
 *	    	 .withCreator("struts", "DogDAO")
 *	    	   .addParam("clas", "com.yourcompany.dao.DogDAO")
 *	    	   .include("getDog")
 *	       .include("getColor")
 *	    	   .exclude("getGender")
 *	    	 .withConverter("dog", "*.Dog")
 *	    	   .addParam("name", "value")
 *		 .withSignature()
 *		   .addLine("import java.util.List;")
 *		   .addLine("import com.example.Check;")
 *		   .addLine("Check.setLotteryResults(List<Integer> nos);")
 *		 .finished();
 *		}
 * }
 *  		</pre>
 *  </ul>
 *  
 * @author Aaron Johnson [ajohnson at cephas dot net / <a href="http://cephas.net/blog">http://cephas.net/blog</a>]
 */
public class FluentConfiguration extends DefaultConfiguration {
	
	/**
	 * typeName and scriptName are used for <allow create .../>
	 */
	private String typeName;
	private String scriptName;
	
	/**
	 * converter and match are used for <allow convert .../>
	 */
	private String converter;
	private String match;
	
	/**
	 * holds name / value pairs used in <allow create|convert ... />
	 */
	private Map params = null;
	
	/**
	 * holds signature lines
	 */
	private StringBuffer signature;
	
	private int CURRENTSTATE = -1;
	private static final int INIT_CREATE = 0;
	private static final int INIT_CONVERT = 1;
	private static final int ALLOW_CREATE = 2;
	private static final int ALLOW_CONVERT = 3;
	private static final int SIGNATURE = 4;
	private static final int COMPLETE = 5;
	
	
	/**
     * The log stream
     */
    private static final Logger log = Logger.getLogger(FluentConfiguration.class);
    
    public FluentConfiguration withConverterType(String id, String className) {
    		setState(INIT_CONVERT);
    		try {
    			Class clazz = Class.forName(className);
    			getConverterManager().addConverterType(id, clazz);
            }
            catch (NoClassDefFoundError ex)
            {
                log.info("Converter '" + id + 
                		"' not loaded due to NoClassDefFoundError. This is only an problem if you wanted to use it. Cause: " + 
                		ex.getMessage());
            }
            catch (ClassNotFoundException ex)
            {
                log.info("Converter '" + id + 
                		"' not loaded due to ClassNotFoundException. This is only an problem if you wanted to use it. Cause: " + 
                		ex.getMessage()); 
            }
            catch (Exception ex)
            {
                log.warn("Failed to load converter '" + id + "', classname=" + className + ": ", ex);
            }
            return this;
    }

	public FluentConfiguration withCreatorType(String id, String className) {
		setState(INIT_CREATE);
		try {
            Class clazz = Class.forName(className);
            getCreatorManager().addCreatorType(id, clazz);
        }
        catch (NoClassDefFoundError ex)
        {
            log.info("Creator '" + id + 
            		"' not loaded due to NoClassDefFoundError. This is only an problem if you wanted to use it. Cause: " + 
            		ex.getMessage());
        }
        catch (ClassNotFoundException ex)
        {
            log.info("Creator '" + id + 
            		"' not loaded due to ClassNotFoundException. This is only an problem if you wanted to use it. Cause: " + 
            		ex.getMessage()); 
        }
        catch (Exception ex)
        {
            log.warn("Failed to load creator '" + id + "', classname=" + className + ": ", ex);
        }
        return this;
	}

	public FluentConfiguration withCreator(String typeName, String scriptName) {
		setState(ALLOW_CREATE);
		this.typeName = typeName;
		this.scriptName = scriptName;
		return this;
	}

	public FluentConfiguration addParam(String name, String value) {
		if (params != null) {
			params.put(name, value);
		} else {
			params = new HashMap();
			params.put(name, value);
		}
		return this; 
	}
	
	public FluentConfiguration include(String methodName) {
		getAccessControl().addIncludeRule(scriptName, methodName);
		return this;
	}
	
	public FluentConfiguration exclude(String methodName) {
		getAccessControl().addExcludeRule(scriptName, methodName);
		return this;
	}
	
	public FluentConfiguration withAuth(String methodName, String role) {
		getAccessControl().addRoleRestriction(scriptName, methodName, role);
		return this;
	}
	
	public FluentConfiguration withConverter(String converter, String match) {
		setState(ALLOW_CONVERT);
		this.converter = converter;
		this.match = match;
		return this;
	}

	public FluentConfiguration withSignature() {
		setState(SIGNATURE);
		return this;
	}
	
	public FluentConfiguration addLine(String line) {
		if (null == line) {
			return this;
		}
		
		if (null == signature) {
			signature = new StringBuffer();
		}
		
		signature.append(line + System.getProperty("line.separator"));
		return this;
	}
	
	/**
	 * Marks the configuration as finished, flushes anything left in the queue.
	 *
	 */
	public void finished() {
		setState(COMPLETE);
	}
	
	/**
	 * Because some parts of the configuration require multiple steps, the instance
	 * needs to maintain a state across invocations. Whenever the state is changed
	 * by calling this method, the instance will 'flush' anything in the queue 
	 * applicable to that state EVEN IF the state itself doesn't change. Thus, it's
	 * important that the child methods don't call setState() when being invoked. 
	 * @param state
	 */
	private void setState(int state) {
		flush();
		CURRENTSTATE = state;
	}
	
	private int getState() {
		return CURRENTSTATE;
	}
	
	private void flush() {
		switch (getState()) {
			case INIT_CONVERT: 
				// do nothing;
				break;
			case INIT_CREATE:
				// do nothing;
				break;
			case ALLOW_CONVERT:
				try {
					getConverterManager().addConverter(match, converter, params);
				} catch (Exception e) {
					log.warn("Failed to add converter of type='" + converter + "', match=" + match + ": ", e);
				}
				params = null;
				match = null;
				converter = null;
				break;
			case ALLOW_CREATE:
				try {
					getCreatorManager().addCreator(typeName, scriptName, params);
				} catch (Exception e) {
					log.warn("Failed to add creator of type='" + typeName + "', scriptName=" + scriptName + ": ", e);
				}
				params = null;
				scriptName = null;
				typeName = null;
				break;
			case SIGNATURE:
				if (signature != null && signature.length() > 0) {
					SignatureParser sigp = new SignatureParser(getConverterManager());
			        sigp.parse(signature.toString());
				}
				break;
			default:
				break;
		}
	}
 }
